/**********************************************************************
 *<
	FILE: asciiout.cpp

	DESCRIPTION:  A utility that outputs an object in ASCII form

	CREATED BY: Rolf Berteig

	HISTORY: created December 26 1995

 *>	Copyright (c) 1994, All Rights Reserved.
 **********************************************************************/
#include "util.h"
#include "utilapi.h"
#include "bmmlib.h"
#include "stdmat.h"
#include "iparamb2.h"
#include <windows.h>
#include <math.h>
#include "ProgressDialog.h"

enum 
{ 
	bmtex_clipu,bmtex_clipv,bmtex_clipw,bmtex_cliph,
	bmtex_jitter,bmtex_usejitter,
	bmtex_apply,bmtex_crop_place,
	bmtex_filtering,
	bmtex_monooutput,
	bmtex_rgboutput,
	bmtex_alphasource,
	bmtex_premultalpha,
	bmtex_bitmap,
	bmtex_coords,	 // access for UVW mapping
	bmtex_output,	 //output window
	bmtex_filename   // bitmap filename virtual parameter
};

enum { multi_params, };  // pblock ID
// multi_params param IDs
enum 
{ 
	multi_mtls,
	multi_ons,
	multi_names,
	multi_ids,

};

#define DIRECTION_RIGHT		2		// material ID 2, is LEFT
#define DIRECTION_LEFT		3		// material ID 3, is RIGHT
#define DIRECTION_TOP		4		// material ID 4, is TOP
#define DIRECTION_BOTTOM	5		// material ID 5, is BOTTOM
#define DIRECTION_FRONT		6		// material ID 6, is FRONT
#define DIRECTION_BACK		7		// material ID 7, is BACK

int GaussianBlurKernel[5][5] = {
	{ 1, 4, 7, 4, 1 },
	{ 4, 16, 26, 16, 4 },
	{ 7, 26,41,26, 7 },
	{ 4, 16, 26, 16, 4 },
	{ 1, 4, 7, 4, 1 }
};

static askedSplineQuestion = FALSE;

Point3 directions[6] = 
{
	Point3(1.0f, 0.0f, 0.0f),		
	Point3(-1.0f, 0.0f, 0.0f),		
	Point3(0.0f, -1.0f, 0.0f),
	Point3(0.0f, 1.0f, 0.0f),
	Point3(0.0f, 0.0f, -1.0f),
	Point3(0.0f, 0.0f, 1.0f)
};

typedef struct POINT3D {
	float X, Y, Z;
	float U, V;
} POINT3D;

typedef struct INTRGBA
{
	int	r;
	int g;
	int b;
	int a;
} INTRGBA;

typedef struct CROPINFO
{
	float	fClipU;
	float	fClipV;
	float	fClipW;
	float	fClipH;
} CROPINFO;

struct gradients {
	gradients( POINT3D const *pVertices );
	float aOneOverZ[3];				// 1/z for each vertex
	float aUOverZ[3];				// u/z for each vertex
	float aVOverZ[3];				// v/z for each vertex
	float dOneOverZdX, dOneOverZdY;	// d(1/z)/dX, d(1/z)/dY
	float dUOverZdX, dUOverZdY;		// d(u/z)/dX, d(u/z)/dY
	float dVOverZdX, dVOverZdY;		// d(v/z)/dX, d(v/z)/dY
};

struct edge {
	edge(gradients const &Gradients,
		POINT3D const *pVertices,
		int Top, int Bottom );
	inline int Step( void );

	float X, XStep;					// fractional x and dX/dY
	int Y, Height;					// current y and vert count
	float OneOverZ, OneOverZStep;		// 1/z and step
	float UOverZ, UOverZStep;			// u/z and step
	float VOverZ, VOverZStep;			// v/z and step
};

void DrawScanLine( Bitmap *pDestInfo, gradients const &Gradients,
	edge *pLeft, edge *pRight, Bitmap *pTextureInfo, WORD alpha);

void TextureMapTriangle( Bitmap *pDestInfo, 
			POINT3D const *pVertices, Bitmap *pTextureInfo, WORD alpha);

INTRGBA *ConvolveImage(INTRGBA *bitmap, int (*kernel)[5][5], int w, int h);
BOOL SplineImages(Bitmap *bitmap1, Bitmap *bitmap2);

#define ASCII_OUT_CLASS_ID		0x4c2b5a6c


BOOL HitChild(INode *node,INode *n)
{
	for (int i =0; i < node->NumberOfChildren(); i++)
		{
		 INode *childNode = node->GetChildNode(i);
		 if (childNode==n) return TRUE;
		 if (HitChild(childNode,n)) return TRUE;
		}
	return FALSE;

}


BOOL ValidateNode(INode *n)
{
//get root node
	
	INode *baseNode = GetCOREInterface()->GetRootNode();
	for (int i =0; i < baseNode->NumberOfChildren(); i++)
		{
		 INode *childNode = baseNode->GetChildNode(i);
		 if (childNode==n) return TRUE;
		 if (HitChild(childNode,n)) return TRUE;
		}
	return FALSE;
//loop through all children looking for match
}


class EdgeList
	{
	public:
		BOOL visible;
		int b;
		Tab<int> connectedFaces;
	};

class EdgeClass
	{
	public:
		Tab<EdgeList*> edges;
		void Free()
			{
			for (int i =0; i < edges.Count(); i++)
				{
				EdgeList *ePtr = edges[i];
				ePtr->connectedFaces.ZeroCount();
				delete ePtr;
				edges[i] = NULL;
				}
			edges.Resize(0);
			};
	};

class UVEdgeClass
	{
public:
	void AddEdge(int a, int b, int faceID, BOOL visible);
	void BuildEdgeList(Mesh *msh, int channel);

	void FreeEdgeList();
	~UVEdgeClass()
		{
		FreeEdgeList();
		}

	Tab<EdgeClass*> e;
	};


void UVEdgeClass::FreeEdgeList()
	{
	for (int i = 0; i < e.Count(); i++)
		{
		EdgeClass *ePtr = e[i];
		ePtr->Free();
		delete ePtr;
		}
	e.Resize(0);
	}

void UVEdgeClass::AddEdge(int a, int b, int faceID, BOOL vis)
	{
	if (e[a] == NULL)
		{
		e[a] = new EdgeClass;
		e[a]->edges.SetCount(1);
		e[a]->edges[0] = new EdgeList;
		e[a]->edges[0]->visible = vis;
		e[a]->edges[0]->b = b;
		e[a]->edges[0]->connectedFaces.Append(1,&faceID);

		}
	else
		{
		EdgeClass *edgePtr = e[a];
		BOOL found = FALSE;
		for (int i =0; i < edgePtr->edges.Count(); i++)
			{
			if (b == edgePtr->edges[i]->b)
				{
				found = TRUE;
				BOOL hitFace = FALSE;
				for (int j=0; j < edgePtr->edges[i]->connectedFaces.Count();j++)
					{
					if (edgePtr->edges[i]->connectedFaces[j] == faceID)
						{
						hitFace = TRUE;
						j = edgePtr->edges[i]->connectedFaces.Count();
						}
					}
				if (!hitFace)
					{
					if ((!edgePtr->edges[i]->visible) && (vis))
						edgePtr->edges[i]->visible = vis;

					edgePtr->edges[i]->connectedFaces.Append(1,&faceID);
					}
				i = edgePtr->edges.Count();
				}
			}
		if (!found)
			{
			EdgeList *el = new EdgeList;
			el-> b = b;
			el->visible = vis;

			el->connectedFaces.Append(1,&faceID);

			edgePtr->edges.Append(1,&el);
			}
		}
	}

void UVEdgeClass::BuildEdgeList(Mesh *msh, int channel)
{
	if (!msh->mapSupport(channel)) return;
//get number texture vertices
	int numberTextureVerts =0;
	numberTextureVerts = msh->getNumMapVerts(channel);
//allocate that many edges
	e.SetCount(numberTextureVerts);
	for (int i =0; i < numberTextureVerts; i++)
		e[i] = NULL;
//get texture face pointer
	int faceCount;
	TVFace *tvFaces=NULL;
	faceCount = msh->numFaces;
	tvFaces = msh->mapFaces(channel);
	Face *faceList =  msh->faces;

	if (tvFaces)
		{
		for (i =0; i < faceCount; i++)
			{
			int a,b;
			for (int j =0; j < 3; j++)
				{
				a = tvFaces->t[j];
				if (j==2)
				b = tvFaces->t[0];
				else b = tvFaces->t[j+1];
				if (a!=b)
					{
					if (a>b)
						{
						int temp;
						temp = a;
						a = b;
						b = temp;
						}
					BOOL vis;
					if (j==0)
						vis = faceList->flags & EDGE_A;
					else if (j==1)
						vis = faceList->flags & EDGE_B;
					else 
						vis = faceList->flags & EDGE_C;

					AddEdge(a,b,i,vis);
					
					}
				}
			
			tvFaces++;
			faceList++;
			}
		}

	}

class AsciiOut : public UtilityObj {
	public:
		BOOL displayOpenEdges;
		int channel;
		ISpinnerControl *Channelspin;

		UVEdgeClass uvEdges;
		
		IUtil *iu;
		Interface *ip;
		HWND hPanel;
		ICustButton *iPick;
		ISpinnerControl *Widthspin;
		ISpinnerControl *Heightspin;
		ISpinnerControl *MatIDspin;
		BOOL m_tmap;

		IColorSwatch *backGroundColor;
		IColorSwatch *lineColor;
		IColorSwatch *tickColor;
		IColorSwatch *openColor;

		Color bColor, lColor, tColor, oColor;
		int mapWidth,mapHeight, matID, mapChannel;
		BOOL useMatID, drawAllEdges,markOpenEdges;
		BOOL useTickMarks, useTextureMaps;
		BOOL useOverlayWire, useSplineMosaic;
		BitmapInfo bi;
		int nTextureMaps;
		HCURSOR waitCursor;
		HCURSOR oldCursor;

		AsciiOut();
		~AsciiOut()
			{
			map->DeleteThis();
			};
		void BeginEditParams(Interface *ip,IUtil *iu);
		void EndEditParams(Interface *ip,IUtil *iu);
		void DeleteThis() {}

		void Init(HWND hWnd);
		void Destroy(HWND hWnd);

		INode *objectNode;
		Bitmap *map;

		void UpdateEdges();

		void SetNode(INode *node);
		void SaveBitmap();
		void UpdateBitmap();

		void RebuildMap();
		Texmap *GetMtlTex(Mtl* mtl, int index);

		void BeginWaitCursor();
		void EndWaitCursor();

		void BXPLine(long x1,long y1,long x2,long y2,
				WORD r, WORD g, WORD b,
				Bitmap *map,
				BOOL wrap);
};

static AsciiOut theAsciiOut;

class AsciiOutClassDesc:public ClassDesc {
	public:
	int 			IsPublic() {return 1;}
	void *			Create(BOOL loading = FALSE) {return &theAsciiOut;}
	const TCHAR *	ClassName() {return GetString(IDS_RB_ASCIIOBJECTOUT);}
	SClass_ID		SuperClassID() {return UTILITY_CLASS_ID;}
	Class_ID		ClassID() {return Class_ID(ASCII_OUT_CLASS_ID,0);}
	const TCHAR* 	Category() {return _T("");}
	};

static AsciiOutClassDesc asciiOutDesc;
ClassDesc* GetAsciiOutDesc() {return &asciiOutDesc;}

class AsciiOutPickNodeCallback : public PickNodeCallback {
	public:		
		BOOL Filter(INode *node);
	};

BOOL AsciiOutPickNodeCallback::Filter(INode *node)
	{
	ObjectState os = node->EvalWorldState(theAsciiOut.ip->GetTime());
	if (os.obj->SuperClassID()==GEOMOBJECT_CLASS_ID &&
		os.obj->IsRenderable()) return TRUE;
	else return FALSE;
	}

static AsciiOutPickNodeCallback thePickFilt;

class AsciiOutPickModeCallback : public PickModeCallback {
	public:		
		BOOL HitTest(IObjParam *ip,HWND hWnd,ViewExp *vpt,IPoint2 m,int flags);
		BOOL Pick(IObjParam *ip,ViewExp *vpt);
		
		void EnterMode(IObjParam *ip) {theAsciiOut.iPick->SetCheck(TRUE);}
		void ExitMode(IObjParam *ip) {theAsciiOut.iPick->SetCheck(FALSE);}

		PickNodeCallback *GetFilter() {return &thePickFilt;}
		BOOL RightClick(IObjParam *ip,ViewExp *vpt) {return TRUE;}

		HWND h;
	};

static AsciiOutPickModeCallback thePickMode;

BOOL AsciiOutPickModeCallback::HitTest(
		IObjParam *ip,HWND hWnd,ViewExp *vpt,IPoint2 m,int flags)
	{
	return ip->PickNode(hWnd,m,&thePickFilt)?TRUE:FALSE;
	}

BOOL AsciiOutPickModeCallback::Pick(IObjParam *ip,ViewExp *vpt)
	{
	INode *node = vpt->GetClosestHit();
	if (node) 
		{
		theAsciiOut.SetNode(node);
		return TRUE;
		}
	else
		{
		return FALSE;
		}			
	}


static BOOL CALLBACK AsciiOutDlgProc(
		HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
	{
	switch (msg) {
		case WM_INITDIALOG:
			theAsciiOut.Init(hWnd);			
			break;
		
		case WM_DESTROY:
			theAsciiOut.Destroy(hWnd);
			break;

		case CC_COLOR_CHANGE:
			switch( LOWORD(wParam) ) 
				{
				case IDC_COLOR_SWATCH_BACKGROUND:
				case IDC_COLOR_SWATCH_LINE:
				case IDC_COLOR_SWATCH_TICK:
				case IDC_COLOR_SWATCH_EDGE:
					if ((theAsciiOut.map != NULL) && (theAsciiOut.map->GetWindow() != NULL))
					{
						theAsciiOut.UpdateBitmap();
					}
					break;
				
				}

			break;
		case WM_CUSTEDIT_ENTER :
			switch( LOWORD(wParam) ) 
				{
				case IDC_WIDTH_EDIT:
				case IDC_HEIGHT_EDIT:
					{
					theAsciiOut.RebuildMap();
					break;
					}
				case IDC_MATID_EDIT:
					if ((theAsciiOut.map != NULL) && (theAsciiOut.map->GetWindow() != NULL))
					{
						theAsciiOut.UpdateBitmap();
					}
					break;
				case IDC_MAP_CHAN:
					{
					theAsciiOut.UpdateEdges();
					if ((theAsciiOut.map != NULL) && (theAsciiOut.map->GetWindow() != NULL))
					{
						theAsciiOut.UpdateBitmap();
					}
					break;
					}

				
				}
			break;

		case CC_SPINNER_BUTTONUP:
			switch( LOWORD(wParam) ) 
				{
				case IDC_MAP_CHAN_SPIN:
					{
					theAsciiOut.UpdateEdges();
					if ((theAsciiOut.map != NULL) && (theAsciiOut.map->GetWindow() != NULL))
					{
						theAsciiOut.UpdateBitmap();
					}
					break;
					}
				case IDC_WIDTH_SPIN:
				case IDC_HEIGHT_SPIN:
					{
					theAsciiOut.RebuildMap();
					break;
					}
				case IDC_MATID_SPIN:
					if ((theAsciiOut.map != NULL) && (theAsciiOut.map->GetWindow() != NULL))
					{
						theAsciiOut.UpdateBitmap();
					}
					break;
				}
			break;

		case WM_COMMAND:
			switch (LOWORD(wParam)) {
/*				case IDOK:
					theAsciiOut.iu->CloseUtility();
					break;				
*/		
				case IDC_MATID_CHECK:
				case IDC_ALLEDGES_CHECK:
				case IDC_OPENDGES_CHECK:
				case IDC_USE_TICKS:					
					if ((theAsciiOut.map != NULL) && (theAsciiOut.map->GetWindow() != NULL))
					{
						theAsciiOut.UpdateBitmap();
					}
					break;
				case IDC_ASCIIOUT_SAVE:
					theAsciiOut.SaveBitmap(); 
					break;

				case IDC_ASCIIOUT_PICK:
					theAsciiOut.ip->SetPickMode(&thePickMode); 
					break;
				case IDC_HELPBUTTON:
					ShellExecute(GetCOREInterface()->GetMAXHWnd(), "open", "iexplore.exe", "www.max3dstuff.com/max4/unwrapUtil/help.html", NULL, SW_SHOWNORMAL);
					break;
				case IDC_TEXTURE_MAP:					
					if (IsDlgButtonChecked(theAsciiOut.hPanel, IDC_TEXTURE_MAP))
					{
						theAsciiOut.useTextureMaps = TRUE;
						CheckDlgButton(theAsciiOut.hPanel, IDC_MATID_CHECK, BST_CHECKED);
						CheckDlgButton(theAsciiOut.hPanel,IDC_OVERLAYWIRE_CHECK, 
								theAsciiOut.useOverlayWire ? BST_CHECKED : BST_UNCHECKED);
						EnableWindow(GetDlgItem(theAsciiOut.hPanel, IDC_MATID_CHECK), FALSE);
						EnableWindow(GetDlgItem(theAsciiOut.hPanel, IDC_OVERLAYWIRE_CHECK), TRUE);
						EnableWindow(GetDlgItem(theAsciiOut.hPanel, IDC_USE_SPLINE), TRUE);
						CheckDlgButton(theAsciiOut.hPanel,IDC_USE_SPLINE, 
							theAsciiOut.useSplineMosaic ? BST_CHECKED : BST_UNCHECKED);
					}
					else
					{
						theAsciiOut.useTextureMaps = FALSE;
						EnableWindow(GetDlgItem(theAsciiOut.hPanel, IDC_MATID_CHECK), TRUE);
						EnableWindow(GetDlgItem(theAsciiOut.hPanel, IDC_OVERLAYWIRE_CHECK), FALSE);
						EnableWindow(GetDlgItem(theAsciiOut.hPanel, IDC_USE_SPLINE), FALSE);
						CheckDlgButton(theAsciiOut.hPanel,IDC_MATID_CHECK, 
								theAsciiOut.useMatID ? BST_CHECKED : BST_UNCHECKED);
					}
					if ((theAsciiOut.map != NULL) && (theAsciiOut.map->GetWindow() != NULL))
					{
						theAsciiOut.UpdateBitmap();
					}
					break;

				case IDC_USE_SPLINE:
					if (IsDlgButtonChecked(theAsciiOut.hPanel, IDC_USE_SPLINE))
					{
						if (!askedSplineQuestion)
						{
							if (MessageBox(NULL, _T("Spline Image Mosiacing can take a few "
									"Minutes, but is far superior in quality\n"
									"Are you sure you want to enable it?\n"
									"This is asked only one time before enabling\n"), 
									_T("Time Warning"),
									MB_OKCANCEL) != IDOK)
								return TRUE;
						}
						askedSplineQuestion = TRUE;
						theAsciiOut.useSplineMosaic = TRUE;
					}
					else
					{
						theAsciiOut.useSplineMosaic = FALSE;
					}

					if ((theAsciiOut.map != NULL) && (theAsciiOut.map->GetWindow() != NULL))
					{
						theAsciiOut.UpdateBitmap();
					}
					break;
				case IDC_OVERLAYWIRE_CHECK:
					theAsciiOut.useOverlayWire = IsDlgButtonChecked(theAsciiOut.hPanel, IDC_OVERLAYWIRE_CHECK);
					if ((theAsciiOut.map != NULL) && (theAsciiOut.map->GetWindow() != NULL))
					{
						theAsciiOut.UpdateBitmap();
					}
					break;

				}
			break;

		default:
			return FALSE;
		}
	return TRUE; 
	}

AsciiOut::AsciiOut()
{
	iu = NULL;
	ip = NULL;	
	hPanel = NULL;	
	iPick = NULL;
	Widthspin = NULL;
	Heightspin = NULL;
	MatIDspin = NULL;
	Channelspin = NULL;

	backGroundColor = NULL;
	lineColor = NULL;
	tickColor = NULL;
	openColor = NULL;
	
	objectNode = NULL;
	map = NULL;

	bColor = Color(0.0f,0.0f,0.0f);
	lColor = Color(1.0f,1.0f,1.0f);
	tColor = Color(1.0f,0.0f,0.0f);
	oColor = Color(0.0f,1.0f,0.0f);
	mapWidth = 640;
	mapHeight = 480;
	matID = 1;
	mapChannel = 1;
	useMatID = FALSE;
	drawAllEdges = FALSE;
	markOpenEdges = TRUE;
	useTickMarks = FALSE;
	useTextureMaps = TRUE;
	useOverlayWire = FALSE;
	useSplineMosaic = FALSE;

	bi.SetWidth(mapWidth);
	bi.SetHeight(mapHeight);
	bi.SetFlags(MAP_HAS_ALPHA);

	waitCursor = LoadCursor(hInstance, IDC_WAIT);
	oldCursor = NULL;
}

void AsciiOut::BeginWaitCursor()
{
	if (waitCursor != NULL)
	{
		oldCursor = SetCursor(waitCursor);
	}
}

void AsciiOut::EndWaitCursor()
{
	if (oldCursor != NULL)
	{
		SetCursor(oldCursor);
		oldCursor = NULL;
	}
}

void AsciiOut::BeginEditParams(Interface *ip,IUtil *iu) 
	{
	this->iu = iu;
	this->ip = ip;
	hPanel = ip->AddRollupPage(
		hInstance,
		MAKEINTRESOURCE(IDD_ASCIIOUT_PANEL),
		AsciiOutDlgProc,
		GetString(IDS_RB_ASCIIOBJECTOUT),
		0);
	}
	
void AsciiOut::EndEditParams(Interface *ip,IUtil *iu) 
	{
	ip->ClearPickMode();
	this->iu = NULL;
	this->ip = NULL;
	ip->DeleteRollupPage(hPanel);
	hPanel = NULL;
	}

void AsciiOut::Init(HWND hWnd)
	{
	iPick = GetICustButton(GetDlgItem(hWnd,IDC_ASCIIOUT_PICK));
	iPick->SetType(CBT_CHECK);
	iPick->SetHighlightColor(GREEN_WASH);
	Widthspin = SetupIntSpinner(hWnd,IDC_WIDTH_SPIN,IDC_WIDTH_EDIT,0,99999999,640);
	Heightspin = SetupIntSpinner(hWnd,IDC_HEIGHT_SPIN,IDC_HEIGHT_EDIT,0,99999999,480);
	MatIDspin = SetupIntSpinner(hWnd,IDC_MATID_SPIN,IDC_MATID_EDIT,0,10000,0);

	Channelspin = SetupIntSpinner(hWnd,IDC_MAP_CHAN_SPIN,IDC_MAP_CHAN,0,99,1);

	backGroundColor = GetIColorSwatch(GetDlgItem(hWnd,IDC_COLOR_SWATCH_BACKGROUND), RGB(255,255,255), _T("Background Color"));
	lineColor = GetIColorSwatch(GetDlgItem(hWnd,IDC_COLOR_SWATCH_LINE), RGB(0,0,0), _T("Line Color"));
	tickColor = GetIColorSwatch(GetDlgItem(hWnd,IDC_COLOR_SWATCH_TICK), RGB(128,0,0), _T("Tick Color"));
	openColor = GetIColorSwatch(GetDlgItem(hWnd,IDC_COLOR_SWATCH_EDGE), RGB(0,200,0), _T("Open Edge Color"));

	m_tmap = FALSE;

	backGroundColor->SetColor(bColor);
	lineColor->SetColor(lColor);
	tickColor->SetColor(tColor);
	openColor->SetColor(oColor);
	

	Widthspin->SetValue(mapWidth,FALSE);
	Heightspin->SetValue(mapHeight,FALSE);
	MatIDspin->SetValue(matID,FALSE);
	Channelspin->SetValue(mapChannel,FALSE);
	

	if (useMatID) CheckDlgButton(hWnd,IDC_MATID_CHECK,BST_CHECKED);
	else CheckDlgButton(hWnd,IDC_MATID_CHECK,BST_UNCHECKED);

	if (drawAllEdges) CheckDlgButton(hWnd,IDC_ALLEDGES_CHECK,BST_CHECKED);
	else CheckDlgButton(hWnd,IDC_ALLEDGES_CHECK,BST_UNCHECKED);

	if (markOpenEdges) CheckDlgButton(hWnd,IDC_OPENDGES_CHECK,BST_CHECKED);
	else CheckDlgButton(hWnd,IDC_OPENDGES_CHECK,BST_UNCHECKED);
	
	if (useTickMarks) CheckDlgButton(hWnd,IDC_USE_TICKS, BST_CHECKED);
	else CheckDlgButton(hWnd,IDC_USE_TICKS, BST_UNCHECKED);

	CheckDlgButton(hWnd,IDC_OVERLAYWIRE_CHECK, useOverlayWire ? BST_CHECKED : BST_UNCHECKED);

	if (useTextureMaps) 
	{
		CheckDlgButton(hWnd,IDC_TEXTURE_MAP,BST_CHECKED);
		CheckDlgButton(hWnd,IDC_MATID_CHECK,BST_CHECKED);
		EnableWindow(GetDlgItem(hWnd, IDC_MATID_CHECK), FALSE);
		EnableWindow(GetDlgItem(hWnd, IDC_OVERLAYWIRE_CHECK), TRUE);
		EnableWindow(GetDlgItem(hWnd, IDC_USE_SPLINE), TRUE);
		CheckDlgButton(hWnd,IDC_USE_SPLINE, useSplineMosaic ? BST_CHECKED : BST_UNCHECKED);
	}
	else 
	{
		CheckDlgButton(hWnd,IDC_MATID_CHECK, useMatID ? BST_CHECKED : BST_UNCHECKED);
		EnableWindow(GetDlgItem(hWnd, IDC_MATID_CHECK), TRUE);
		EnableWindow(GetDlgItem(hWnd, IDC_OVERLAYWIRE_CHECK), FALSE);
		EnableWindow(GetDlgItem(hWnd, IDC_USE_SPLINE), FALSE);
	}


	if (!ValidateNode(objectNode))
		objectNode = NULL;

//validate node
	if (objectNode)
		iPick->SetText(objectNode->GetName());
	CheckRadioButton( hPanel,IDC_MAP_TEXMAP,IDC_MAP_VERTCOL,IDC_MAP_TEXMAP);
 

	}




void AsciiOut::Destroy(HWND hWnd)
	{

	bColor = backGroundColor->GetColor();
	lColor = lineColor->GetColor();
	tColor = tickColor->GetColor();
	oColor = openColor->GetColor();
	

	mapWidth = Widthspin->GetIVal();
	mapHeight = Heightspin->GetIVal();
	matID = MatIDspin->GetIVal();
	mapChannel = Channelspin->GetIVal();
	


	useMatID= IsDlgButtonChecked(hPanel,IDC_MATID_CHECK);
	drawAllEdges= IsDlgButtonChecked(hPanel,IDC_ALLEDGES_CHECK);
	markOpenEdges= IsDlgButtonChecked(hPanel,IDC_OPENDGES_CHECK);


	ReleaseISpinner(Widthspin);
	ReleaseISpinner(Heightspin);
	ReleaseISpinner(MatIDspin);
	ReleaseISpinner(Channelspin);

	ReleaseICustButton(iPick);
	ReleaseIColorSwatch(backGroundColor);
	ReleaseIColorSwatch(lineColor);
	ReleaseIColorSwatch(tickColor);
	ReleaseIColorSwatch(openColor);
	iPick = NULL;
	}

class NullView: public View {
	public:
		Point2 ViewToScreen(Point3 p) { return Point2(p.x,p.y); }
		NullView() { worldToView.IdentityMatrix(); screenW=640.0f; screenH = 480.0f; }
	};


void AsciiOut::RebuildMap()
	{
	if (!ValidateNode(objectNode))
		objectNode = NULL;

	if (objectNode)
		{
		int w,h;
		w = Widthspin->GetIVal();
		h = Heightspin->GetIVal();
		if ((w != bi.Width()) || (h!=bi.Height()))
			{
			bi.SetWidth(Widthspin->GetIVal());
			bi.SetHeight(Heightspin->GetIVal());
			map->UnDisplay();
			map->DeleteThis();
			map = NULL;
			UpdateBitmap();
			if (map) map->Display();
			}
		}

	}

void AsciiOut::UpdateEdges()
	{
	if (!ValidateNode(objectNode))
		objectNode = NULL;

	if (objectNode)
		{
		ObjectState os = objectNode->EvalWorldState(theAsciiOut.ip->GetTime());
		assert(os.obj->SuperClassID()==GEOMOBJECT_CLASS_ID);
		BOOL needDel;
		NullView nullView;
		Mesh *mesh = ((GeomObject*)os.obj)->GetRenderMesh(ip->GetTime(),objectNode,nullView,needDel);
		if (!mesh) return;
	
		channel = Channelspin->GetIVal();

		uvEdges.BuildEdgeList(mesh, channel);

		if (needDel) delete mesh;
		}
	}

void AsciiOut::SetNode(INode *node)
{
	DWORD flags;
	objectNode = node;
	bi.SetWidth(Widthspin->GetIVal());
	bi.SetHeight(Heightspin->GetIVal());
	bi.SetType(BMM_TRUE_64);
	flags = bi.Flags();
	flags &= ~MAP_ALPHA_PREMULTIPLIED;
	flags |= MAP_HAS_ALPHA;
	bi.SetFlags(flags);
	iPick->SetText(node->GetName());

	UpdateEdges();

	UpdateBitmap();

	if (map) map->Display();
}

void AsciiOut::SaveBitmap()
	{
	UpdateBitmap();
	BOOL iret = TheManager->SelectFileOutput(&bi ,hPanel ,"Select File");
	if (iret)
		{
		map->OpenOutput(&bi);
		map->Write(&bi);
		map->Close(&bi);
		}

	}

void AsciiOut::UpdateBitmap()
{
	BOOL bUseTextures;
	TVFace *TVFacePreMap[7];
    BitmapTex *pTexMap;
	Bitmap *mapSrc[7];
	int mapChannels[7];
	int matIDS[7];
	CROPINFO rgCropInfo[7];
	int i, j, k;
	BOOL bMulti;
	int nMaterials;
	IParamBlock2 *pParamBlock;
	TCHAR message[256];
	BOOL useMatIDs;
	Bitmap *tImages[7];
	BOOL bFirstFound;
	int saveNTextures;

	//get object	
	INode *node=NULL;

	int channel = Channelspin->GetIVal();

	bFirstFound = FALSE;
	if (!ValidateNode(objectNode))
		objectNode = NULL;

	if (objectNode == NULL) 
		return;
	else 
		node = objectNode;

	if (map == NULL)
		map = TheManager->Create(&bi);

	BMM_Color_64 bit;
	Point3 VertA,VertB,VertC;
	int MatID1,MatID2;

	bit.r = 0;
	bit.g = 0;
	bit.b = 0;
	bit.a = 0xFFFF;

	COLORREF color = backGroundColor->GetColor();
	int r = (int) GetRValue(color)*512;
	int g = (int) GetGValue(color)*512;
	int b = (int) GetBValue(color)*512;


	map->Fill(r,g,b,0);


	color = lineColor->GetColor();
	int lr = (int) GetRValue(color)*512;
	int lg = (int) GetGValue(color)*512;
	int lb = (int) GetBValue(color)*512;


	color = tickColor->GetColor();
	int tr = (int) GetRValue(color)*512;
	int tg = (int) GetGValue(color)*512;
	int tb = (int) GetBValue(color)*512;

	color = openColor->GetColor();
	int or = (int) GetRValue(color)*512;
	int og = (int) GetGValue(color)*512;
	int ob = (int) GetBValue(color)*512;


	ObjectState os = node->EvalWorldState(theAsciiOut.ip->GetTime());
	assert(os.obj->SuperClassID()==GEOMOBJECT_CLASS_ID);
	BOOL needDel;
	NullView nullView;
	Mesh *mesh = ((GeomObject*)os.obj)->GetRenderMesh(ip->GetTime(),node,nullView,needDel);
	if (!mesh) return;
	if (!mesh->mapSupport(channel)) return;

	Matrix3 tm = node->GetObjTMAfterWSM(theAsciiOut.ip->GetTime());

	BitArray sel = mesh->VertSel();
	Face *FaceList =  mesh->faces;
	TVFace	*TVFaceList = mesh->mapFaces(channel);
	UVVert *tverts = mesh->mapVerts(channel);

	bUseTextures = useTextureMaps;
	// Kenneth Hurley, additions added here
	if (bUseTextures)
	{
		Mtl* mtl = node->GetMtl();
		StdMat* sm;
	    if (!mtl)
			return;

		// We only handle standard materials and multi-subobject
		if (mtl->ClassID() == Class_ID(DMTL_CLASS_ID, 0))
		{
			sm = (StdMat*) mtl;
			nMaterials = 1;
			bMulti = FALSE;
		}
		else if (mtl->ClassID() == Class_ID(MULTI_CLASS_ID, 0))
		{
			nMaterials = mtl->NumSubMtls();
			bMulti = TRUE;
		}

		nTextureMaps = 0;
		for (i=0; i<nMaterials; i++)
		{
			if (bMulti)
			{
				sm = (StdMat*) mtl->GetSubMtl(i);
			}

			// Check for diffuse texture map
			pTexMap = (BitmapTex*) sm->GetSubTexmap(ID_DI);
			if ((pTexMap != NULL) && (pTexMap->ClassID() == Class_ID(BMTEX_CLASS_ID, 0x00)))
			{
				if (sm && (i < DIRECTION_RIGHT-1) || (i > DIRECTION_BACK-1))
				{
					if (bFirstFound)
					{
						continue;
					}

					saveNTextures = nTextureMaps;
					nTextureMaps = 6;
					bFirstFound = TRUE;
				}

				if (pTexMap->GetUVWSource() != UVWSRC_EXPLICIT)
				{
					MessageBox(NULL, _T("UVW mapping must be set to explicit, aborting.."),
						_T("Map Error!"), MB_OK);
					return;
				}

				matIDS[nTextureMaps] = i+1;
				mapChannels[nTextureMaps] = pTexMap->GetMapChannel();
				// don't include this texture map
				if (!mesh->mapSupport(mapChannels[nTextureMaps]) && 
					(sm && (i >= DIRECTION_RIGHT-1) && (i <= DIRECTION_BACK-1)))
				{
					wsprintf(message, "There are no faces with map channel %d.\n"
								"Ignoring texture map id %d.\n", mapChannels[nTextureMaps],
								i+1);
					MessageBox(NULL, message, _T("Map Channel Warning!"), MB_OK);
					break;
				}
				TVFacePreMap[nTextureMaps] = mesh->mapFaces(mapChannels[nTextureMaps]);
				// grab material applied to this node

				mapSrc[nTextureMaps] = pTexMap->GetBitmap(0);


				rgCropInfo[nTextureMaps].fClipU = 0.0f;
				rgCropInfo[nTextureMaps].fClipV = 0.0f;
				rgCropInfo[nTextureMaps].fClipW = 1.0f;
				rgCropInfo[nTextureMaps].fClipH = 1.0f;

				pParamBlock = pTexMap->GetParamBlock(0);
				if (pParamBlock != NULL)
				{
					pParamBlock->GetValue(bmtex_clipu, 0, rgCropInfo[nTextureMaps].fClipU, FOREVER);
					pParamBlock->GetValue(bmtex_clipv, 0, rgCropInfo[nTextureMaps].fClipV, FOREVER);
					pParamBlock->GetValue(bmtex_clipw, 0, rgCropInfo[nTextureMaps].fClipW, FOREVER);
					pParamBlock->GetValue(bmtex_cliph, 0, rgCropInfo[nTextureMaps].fClipH, FOREVER);       
				}
				// material ID must match these for this to work
				if (sm && (i >= DIRECTION_RIGHT-1) && (i <= DIRECTION_BACK-1))
				{
					nTextureMaps++;
				}
				else
				{
					nTextureMaps = saveNTextures;
				}
			}
		}
	}

	if (bFirstFound && (nTextureMaps == 0))
	{
		mapChannels[0] = mapChannels[6];
		TVFacePreMap[0] = TVFacePreMap[6];
		mapSrc[0] = mapSrc[6];
		matIDS[0] = 0;
		rgCropInfo[0] = rgCropInfo[6];
		if (!mesh->mapSupport(mapChannels[0]))
		{
			wsprintf(message, "There are no faces with map channel %d.\n"
						"Ignoring texture map.\n", mapChannels[0],
						i+1);
			MessageBox(NULL, message, _T("Map Channel Warning!"), MB_OK);
		}
		else
		{
			nTextureMaps = 1;
		}
	}

	if (nTextureMaps == 0)
	{
		bUseTextures = FALSE;
	}

	MatID1 = MatIDspin->GetIVal();
	if (bUseTextures)
	{
		POINT3D Vert[3];
		Point3 fn;
		Box3 bbox = mesh->getBoundingBox(&tm);
		Point3 bboxSize = bbox.Width();

		if (bboxSize.x == 0.0f)
		{
			bboxSize.x = 0.00000000001f;
		}

		if (bboxSize.y == 0.0f)
		{
			bboxSize.y = 0.00000000001f;
		}

		if (bboxSize.z == 0.0f)
		{
			bboxSize.z = 0.00000000001f;
		}

		// create an image with mask of this direction
		for (k=0; k<nTextureMaps; k++)
		{
			if (k == 0)
			{
				tImages[0] = map;
			}
			else
			{
				if (useSplineMosaic)
					tImages[k] = TheManager->Create(&bi);
				else
					tImages[k] = map;
			}

			if (tImages[k] == NULL)
			{
				MessageBox(NULL, _T("Unable to create temporary bitmap for splining.\n"),
					_T("Memory Error!"), MB_OK);
				return;
			}

			if (mesh->normalsBuilt == FALSE)
				mesh->buildNormals();

			for (i=0; i<mesh->getNumFaces(); i++) 
			{
				int nFace;
				int index;

				fn = tm.VectorTransform(mesh->getFaceNormal(i));
				float temp = fn.z;
				fn.z = fn.y;
				fn.y = temp;

				// find mostly direction facing
				float mostFace = -1.0f;
				nFace = 0;
				index = k;
				for (j=0; j<nTextureMaps; j++)
				{
					float dotP;
					dotP = DotProd(directions[matIDS[j]-DIRECTION_RIGHT], fn);
					if (dotP > mostFace)
					{
						mostFace = dotP;
						nFace = j;
					}

				}

				if (!useSplineMosaic)
				{
					index = nFace;
				}

				UVVert *srcUV = mesh->mapVerts(mapChannels[index]);
				if  (TVFaceList && (mapSrc[index] != NULL))
				{
					float fX, fY, fZ;
					for (j=0; j<3; j++)
					{
						Vert[j].X = tverts[TVFaceList[i].t[j]].x;
						Vert[j].Y = tverts[TVFaceList[i].t[j]].y;
						Vert[j].Z = tverts[TVFaceList[i].t[j]].z;

						fZ = srcUV[(TVFacePreMap[index])[i].t[j]].z;
						fX = srcUV[(TVFacePreMap[index])[i].t[j]].x;
						fY = 1.0f - srcUV[(TVFacePreMap[index])[i].t[j]].y;
						fX = (fX > 1.0f) ? 1.0f : ((fX < 0.0f) ? 0.0f : fX);
						fY = (fY > 1.0f) ? 1.0f : ((fY < 0.0f) ? 0.0f : fY);
						fX = (fX * rgCropInfo[index].fClipW) + rgCropInfo[index].fClipU;
						fY = (fY * rgCropInfo[index].fClipH) + rgCropInfo[index].fClipV;
						Vert[j].U = fX;
						Vert[j].V = fY;
						// move to destination texture space
						Vert[j].X = Vert[j].X * map->Width();
						// clamp
						Vert[j].X = min(Vert[j].X, map->Width()-1);
						Vert[j].X = max(Vert[j].X, 0);

						// move to source texture space
						Vert[j].U = min(Vert[j].U, 1.0f);
						Vert[j].U = max(Vert[j].U, 0);

						Vert[j].Y = Vert[j].Y * map->Height();
						// clamp
						Vert[j].Y = min(Vert[j].Y, map->Height());
						Vert[j].Y = max(Vert[j].Y, 0);

						Vert[j].V = min(Vert[j].V, 1.0f);
						Vert[j].V = max(Vert[j].V, 0);

					}
					TextureMapTriangle(tImages[k], Vert, mapSrc[index], (nFace == index) ? 65535 : 0); 
				 }
			}
		}

//	theProgress = new CProgress(hPanel);

		if (useSplineMosaic)
		{
			// now spline the images together
			for (k=1; k<nTextureMaps; k++)
			{
				BeginWaitCursor();
				if (!SplineImages(map, tImages[k]))
				{
					for (j=k; j<nTextureMaps; j++)
						tImages[j]->DeleteThis();
					return;
				} 
				tImages[k]->DeleteThis();
				EndWaitCursor();
			}
		}
	}

//	theProgress->Close();
//	delete theProgress;

	useMatIDs = IsDlgButtonChecked(hPanel,IDC_MATID_CHECK);

//draw lines
	if (!bUseTextures || IsDlgButtonChecked(hPanel,IDC_OVERLAYWIRE_CHECK))
	{
		for (i=0; i< uvEdges.e.Count(); i++) 
			{
			int a,b;
			a = i;
			if (uvEdges.e[i])
				{
				for (int j = 0; j < uvEdges.e[i]->edges.Count(); j++)
					{
					BOOL draw = FALSE;
					if (useMatIDs && !bUseTextures)
						{
						for (int k = 0; k < uvEdges.e[i]->edges[j]->connectedFaces.Count(); k++)
							{
							 
							MatID2 = FaceList[uvEdges.e[i]->edges[j]->connectedFaces[k]].getMatID()+1;;
							if (MatID2 == MatID1)
								{
								draw = TRUE;
								k = uvEdges.e[i]->edges[j]->connectedFaces.Count();
								}
							}
						}
					else draw = TRUE;
					 
					if (( (uvEdges.e[i]->edges[j]->visible) || (IsDlgButtonChecked(hPanel,IDC_ALLEDGES_CHECK)) ) && (draw))
						{
						b = uvEdges.e[i]->edges[j]->b;
						VertA = tverts[a];
						VertB = tverts[b];
						if ( (!IsDlgButtonChecked(hPanel,IDC_OPENDGES_CHECK)) ||
						   (uvEdges.e[i]->edges[j]->connectedFaces.Count() > 1) )
							BXPLine((long) (map->Width()*VertA.x),map->Height()-1-(long) ((map->Height()-1)*VertA.y),
									(long) (map->Width()*VertB.x),map->Height()-1-(long) ((map->Height()-1)*VertB.y),
								   lr, lg, lb,
								map,TRUE);
						else
							BXPLine((long) (map->Width()*VertA.x),map->Height()-1-(long) ((map->Height()-1)*VertA.y),
									(long) (map->Width()*VertB.x),map->Height()-1-(long) ((map->Height()-1)*VertB.y),
								   or, og, ob,
								map,TRUE);
						}

					}
				}
			}
	}

	BOOL useTickMarks = IsDlgButtonChecked(hPanel,IDC_USE_TICKS);
	if (useTickMarks)
	{
//put in ticks

		for (i=0; i<mesh->getNumFaces(); i++) 
		{
			if  (TVFaceList)   		  
			{
	//get tvverst
				 MatID2 = FaceList->getMatID()+1;
				 if ( (!useMatIDs) ||
					  (MatID1 == MatID2)    )
				 {				 
				 VertA = (Point3) tverts[TVFaceList->t[0]];
				 VertB = (Point3) tverts[TVFaceList->t[1]];
				 VertC = (Point3) tverts[TVFaceList->t[2]];

	/*			 map->PutPixels((int)(map->Width()*VertA.x),map->Height()-1-(int)((map->Height()-1)*VertA.y),1,&bit);
				 map->PutPixels((int)(map->Width()*VertB.x),map->Height()-1-(int)((map->Height()-1)*VertB.y),1,&bit);
				 map->PutPixels((int)(map->Width()*VertC.x),map->Height()-1-(int)((map->Height()-1)*VertC.y),1,&bit);

	//Draw Lines from 0A to 1B
				 if ( (FaceList->flags & EDGE_A) || (IsDlgButtonChecked(hPanel,IDC_ALLEDGES_CHECK)) )
					{
					BXPLine((long) (map->Width()*VertA.x),map->Height()-1-(long) ((map->Height()-1)*VertA.y),
							(long) (map->Width()*VertB.x),map->Height()-1-(long) ((map->Height()-1)*VertB.y),
						   lr, lg, lb,
						   map,TRUE);

					}
	//Draw Lines from 1B to 2C
				 if ( (FaceList->flags & EDGE_B) || (IsDlgButtonChecked(hPanel,IDC_ALLEDGES_CHECK)) )
					{
					BXPLine((long) (map->Width()*VertB.x),map->Height()-1-(long) ((map->Height()-1)*VertB.y),
							(long) (map->Width()*VertC.x),map->Height()-1-(long) ((map->Height()-1)*VertC.y),
						   lr, lg, lb,
						   map,TRUE);

					}
	//Draw Lines from 1A to 2C
				 if ( (FaceList->flags & EDGE_C) || (IsDlgButtonChecked(hPanel,IDC_ALLEDGES_CHECK)) )
					{
					BXPLine((long) (map->Width()*VertA.x),map->Height()-1-(long) ((map->Height()-1)*VertA.y),
							(long) (map->Width()*VertC.x),map->Height()-1-(long) ((map->Height()-1)*VertC.y),
						   lr, lg, lb,
						   map,TRUE);

					}
	*/

				 
				 
				 if (sel[FaceList->v[0]])
					{
	//draw red tick

					BXPLine((long) (map->Width()*VertA.x)-4,map->Height()-1-(long) ((map->Height()-1)*VertA.y)-4,
							(long) (map->Width()*VertA.x)+4,map->Height()-1-(long) ((map->Height()-1)*VertA.y)+4,
						   tr, tg, tb,
						   map,FALSE);
					BXPLine((long) (map->Width()*VertA.x)+4,map->Height()-1-(long) ((map->Height()-1)*VertA.y)-4,
							(long) (map->Width()*VertA.x)-4,map->Height()-1-(long) ((map->Height()-1)*VertA.y)+4,
						   tr, tg, tb,
						   map,FALSE);

				 }
				  if (sel[FaceList->v[1]])
					{
	//draw red tick
					BXPLine((long) (map->Width()*VertB.x)-4,map->Height()-1-(long) ((map->Height()-1)*VertB.y)-4,
							(long) (map->Width()*VertB.x)+4,map->Height()-1-(long) ((map->Height()-1)*VertB.y)+4,
						   tr, tg, tb,
						   map,FALSE);
					BXPLine((long) (map->Width()*VertB.x)+4,map->Height()-1-(long) ((map->Height()-1)*VertB.y)-4,
							(long) (map->Width()*VertB.x)-4,map->Height()-1-(long) ((map->Height()-1)*VertB.y)+4,
						   tr, tg, tb,
						   map,FALSE);
					}

				if (sel[FaceList->v[2]])
					{
	//draw red tick
					BXPLine((long) (map->Width()*VertC.x)-4,map->Height()-1-(long) ((map->Height()-1)*VertC.y)-4,
							(long) (map->Width()*VertC.x)+4,map->Height()-1-(long) ((map->Height()-1)*VertC.y)+4,
						   tr, tg, tb,
						   map,FALSE);
					BXPLine((long) (map->Width()*VertC.x)+4,map->Height()-1-(long) ((map->Height()-1)*VertC.y)-4,
							(long) (map->Width()*VertC.x)-4,map->Height()-1-(long) ((map->Height()-1)*VertC.y)+4,
						   tr, tg, tb,
						   map,FALSE);
					}

				}

			 }
			 FaceList++;
			 TVFaceList++;

		}
	}

	if (needDel) delete mesh;

	map->RefreshWindow();

}

void AsciiOut::BXPLine(long x1,long y1,long x2,long y2,
					   WORD r, WORD g, WORD b,
					   Bitmap *map, BOOL wrap)



{
  long i,px,py,x,y;
  long dx,dy,dxabs,dyabs,sdx,sdy;
  long count;

  BMM_Color_64 bit;

  bit.r = r;
  bit.g = g;
  bit.b = b;
  bit.a = 0xFFFF;

/*
  if (x1<0) x1= 0;
  if (x2<0) x2= 0;
  if (y1<0) y1= 0;
  if (y2<0) y2= 0;

  if (x1>=map->GetWidth()) x1= map->GetWidth()-1;
  if (x2>=map->GetWidth()) x2= map->GetWidth()-1;
  if (y1>=h) y1= h-1;
  if (y2>=h) y2= h-1;
 */

 dx = x2-x1;
 dy = y2-y1;

 if (dx >0)
	 sdx = 1;
	else
	 sdx = -1;

 if (dy >0)
	 sdy = 1;
	else
	 sdy = -1;


 dxabs = abs(dx);
 dyabs = abs(dy);

 x=0;
 y=0;
 px=x1;
 py=y1;

 count = 0;

 if (dxabs >= dyabs)
	 for (i=0; i<=dxabs; i++)
		{
		 y +=dyabs;
		 if (y>=dxabs)
			{
			 y-=dxabs;
			 py+=sdy;
			}
		if (wrap)
			{
			 if (px>=map->Width())
				{
				px = 0;
				}
			 if (px<0)
				{
				px = map->Width()-1;
				}
			 if (py>=map->Height())
				{
				py = 0;
				}
			 if (py<0)
				{
				px = map->Height()-1;
				}
			map->PutPixels(px,py,1,&bit);
			}
			else
			{
			 if ( (px>=0) && (px<map->Width()) &&
				  (py>=0) && (py<map->Height())    )
				{
				map->PutPixels(px,py,1,&bit);
				}

			}
		 px+=sdx;
		 }


		 else

	 for (i=0; i<=dyabs; i++)
		{
		 x+=dxabs;
		 if (x>=dyabs)
			{
			 x-=dyabs;
			 px+=sdx;
			}
		if (wrap)
			{
			 if (px>=map->Width())
				{
				px = 0;
				}
			 if (px<0)
				{
				px = map->Width()-1;
				}
			 if (py>=map->Height())
				{
				py = 0;
				}
			 if (py<0)
				{
				px = map->Height()-1;
				}
			map->PutPixels(px,py,1,&bit);
			}
			else
			{
			 if ( (px>=0) && (px<map->Width()) &&
				  (py>=0) && (py<map->Height())    )
				{
				map->PutPixels(px,py,1,&bit);
				}

			}

		 map->PutPixels(px,py,1,&bit);
		 py+=sdy;
		}
(count)--;
}

inline int edge::Step( void ) {
	X += XStep; Y++; Height--;
	UOverZ += UOverZStep; VOverZ += VOverZStep;
	OneOverZ += OneOverZStep;
	return Height;
}

/******** TextureMapTriangle **********/
// Courtesy of Chris Hecker 
// Fixes by Kenneth Hurley

void TextureMapTriangle( Bitmap *pDestInfo, 
			POINT3D const *pVertices, Bitmap *pTextureInfo, WORD alpha)
{
	int Top, Middle, Bottom;
	int MiddleCompare, BottomCompare;
	float Y0 = pVertices[0].Y;
	float Y1 = pVertices[1].Y;
	float Y2 = pVertices[2].Y;


	// sort vertices in y
	if(Y0 < Y1) {
		if(Y2 < Y0) {
			Top = 2; Middle = 0; Bottom = 1;
			MiddleCompare = 0; BottomCompare = 1;
		} else {
			Top = 0;
			if(Y1 < Y2) {
				Middle = 1; Bottom = 2;
				MiddleCompare = 1; BottomCompare = 2;
			} else {
				Middle = 2; Bottom = 1;
				MiddleCompare = 2; BottomCompare = 1;
			}
		}
	} else {
		if(Y2 < Y1) {
			Top = 2; Middle = 1; Bottom = 0;
			MiddleCompare = 1; BottomCompare = 0;
		} else {
			Top = 1;
			if(Y0 < Y2) {
				Middle = 0; Bottom = 2;
				MiddleCompare = 3; BottomCompare = 2;
			} else {
				Middle = 2; Bottom = 0;
				MiddleCompare = 2; BottomCompare = 3;
			}
		}
	}

	gradients Gradients(pVertices);
	edge TopToBottom(Gradients,pVertices,Top,Bottom);
	edge TopToMiddle(Gradients,pVertices,Top,Middle);
	edge MiddleToBottom(Gradients,pVertices,Middle,Bottom);
	edge *pLeft, *pRight;
	int MiddleIsLeft;

	// the triangle is clockwise, so
	// if bottom > middle then middle is right
	if(BottomCompare > MiddleCompare) {
		MiddleIsLeft = 0;
		pLeft = &TopToBottom; pRight = &TopToMiddle;
	} else {
		MiddleIsLeft = 1;
		pLeft = &TopToMiddle; pRight = &TopToBottom;
	}

	int Height = TopToMiddle.Height;

	while(Height--) {
		DrawScanLine(pDestInfo, Gradients, pLeft, pRight, pTextureInfo, alpha);
		TopToMiddle.Step(); TopToBottom.Step();
	}

	Height = MiddleToBottom.Height;

	if(MiddleIsLeft) {
		pLeft = &MiddleToBottom; pRight = &TopToBottom;
	} else {
		pLeft = &TopToBottom; pRight = &MiddleToBottom;
	}
	
	while(Height--) {
		DrawScanLine(pDestInfo, Gradients, pLeft, pRight, pTextureInfo, alpha);
		MiddleToBottom.Step(); TopToBottom.Step();
	}
}

/********** gradients constructor **********/

gradients::gradients( POINT3D const *pVertices )
{
	int Counter;
	
	float OneOverdX = 1 /(((pVertices[1].X - pVertices[2].X) *
			(pVertices[0].Y - pVertices[2].Y)) -
			((pVertices[0].X - pVertices[2].X) *
			(pVertices[1].Y - pVertices[2].Y)));

	float OneOverdY = -OneOverdX;

	for(Counter = 0;Counter < 3;Counter++) {
		float const OneOverZ = 1/pVertices[Counter].Z;
		aOneOverZ[Counter] = OneOverZ;
		aUOverZ[Counter] = pVertices[Counter].U * OneOverZ;
		aVOverZ[Counter] = pVertices[Counter].V * OneOverZ;
	}

	dOneOverZdX = OneOverdX * (((aOneOverZ[1] - aOneOverZ[2]) *
			(pVertices[0].Y - pVertices[2].Y)) -
			((aOneOverZ[0] - aOneOverZ[2]) *
			(pVertices[1].Y - pVertices[2].Y)));
	dOneOverZdY = OneOverdY * (((aOneOverZ[1] - aOneOverZ[2]) *
			(pVertices[0].X - pVertices[2].X)) -
			((aOneOverZ[0] - aOneOverZ[2]) *
			(pVertices[1].X - pVertices[2].X)));

	dUOverZdX = OneOverdX * (((aUOverZ[1] - aUOverZ[2]) *
			(pVertices[0].Y - pVertices[2].Y)) -
			((aUOverZ[0] - aUOverZ[2]) *
			(pVertices[1].Y - pVertices[2].Y)));
	dUOverZdY = OneOverdY * (((aUOverZ[1] - aUOverZ[2]) *
			(pVertices[0].X - pVertices[2].X)) -
			((aUOverZ[0] - aUOverZ[2]) *
			(pVertices[1].X - pVertices[2].X)));

	dVOverZdX = OneOverdX * (((aVOverZ[1] - aVOverZ[2]) *
			(pVertices[0].Y - pVertices[2].Y)) -
			((aVOverZ[0] - aVOverZ[2]) *
			(pVertices[1].Y - pVertices[2].Y)));
	dVOverZdY = OneOverdY * (((aVOverZ[1] - aVOverZ[2]) *
			(pVertices[0].X - pVertices[2].X)) -
			((aVOverZ[0] - aVOverZ[2]) *
			(pVertices[1].X - pVertices[2].X)));
}

/********** edge constructor ***********/

edge::edge( gradients const &Gradients,
		POINT3D const *pVertices, int Top, int Bottom )
{
	Y = ceil(pVertices[Top].Y);
	int YEnd = ceil(pVertices[Bottom].Y);
	Height = YEnd - Y;

	float YPrestep = Y - pVertices[Top].Y;
	
	float RealHeight = pVertices[Bottom].Y - pVertices[Top].Y;
	float RealWidth = pVertices[Bottom].X - pVertices[Top].X;

	X = ((RealWidth * YPrestep)/RealHeight) + pVertices[Top].X;
	XStep = RealWidth/RealHeight;
	float XPrestep = X - pVertices[Top].X;

	OneOverZ = Gradients.aOneOverZ[Top] +
		YPrestep * Gradients.dOneOverZdY +
		XPrestep * Gradients.dOneOverZdX;
	OneOverZStep = XStep *
		Gradients.dOneOverZdX + Gradients.dOneOverZdY;

	UOverZ = Gradients.aUOverZ[Top] +
		YPrestep * Gradients.dUOverZdY +
		XPrestep * Gradients.dUOverZdX;
	UOverZStep = XStep *
		Gradients.dUOverZdX + Gradients.dUOverZdY;

	VOverZ = Gradients.aVOverZ[Top] +
		YPrestep * Gradients.dVOverZdY +
		XPrestep * Gradients.dVOverZdX;
	VOverZStep = XStep *
		Gradients.dVOverZdX + Gradients.dVOverZdY;
}

/********** DrawScanLine ************/

void DrawScanLine( Bitmap *pDestInfo,
		gradients const &Gradients,
		edge *pLeft, edge *pRight,
		Bitmap *pTextureInfo, WORD alpha)
{
	BMM_Color_64 color;

	// we assume dest and texture are top-down

	int XStart = ceil(pLeft->X);
	float XPrestep = XStart - pLeft->X;

	int Width = ceil(pRight->X) - XStart;

	float OneOverZ = pLeft->OneOverZ +
			XPrestep * Gradients.dOneOverZdX;
	float UOverZ = pLeft->UOverZ +
			XPrestep * Gradients.dUOverZdX;
	float VOverZ = pLeft->VOverZ +
			XPrestep * Gradients.dVOverZdX;

	if(Width > 0) {
		while (Width--) {
			float Z = 1/OneOverZ;
			int U = (UOverZ * Z * pTextureInfo->Width()) - .5f;
			int V = (VOverZ * Z * pTextureInfo->Height()) - .5f;

			pTextureInfo->GetPixels(U, V, 1, &color);
			color.a = alpha;
			pDestInfo->PutPixels(XStart, pDestInfo->Height() - 1 - pLeft->Y, 1, &color);
			XStart++;
	
			OneOverZ += Gradients.dOneOverZdX;
			UOverZ += Gradients.dUOverZdX;
			VOverZ += Gradients.dVOverZdX;

		}
	}
}

#define BORDERSIZE 4

//=====================================================================
// Function:	CreateBorderImage
// Description:	Create a copy of the bitmap with a 2 pixel border
// Parameters:	Bitmap = 64 bit TRUE color bitmap pointer
//				w = width of image
//				h = height of image
// Returns:		pointer to new Bitmap class or NULL for error
//=====================================================================
INTRGBA *CreateBorderImage(INTRGBA *bitmap, int w, int h)
{
	INTRGBA *src;
	INTRGBA *dest;
	int x, y;
	INTRGBA *temp;
	
	temp = new INTRGBA[(w+BORDERSIZE)*(h+BORDERSIZE)];
	if (temp == NULL)
	{
		return NULL;
	}

	dest = (INTRGBA *)temp + ((w + BORDERSIZE) * 2) + (BORDERSIZE/2);

	for (y=0; y<h; y++)
	{
		memcpy(dest, bitmap, (w * sizeof(INTRGBA)));
		dest += (w + BORDERSIZE);
		bitmap += w;
	}

	src = dest = temp + ((w + BORDERSIZE) * 2);
	src += BORDERSIZE/2;

	// copy reflect pixels around x edge
	for (y=0; y<h; y++)
	{
		// left -2
		dest->r = 2 * src->r - (src+2)->r;
		dest->g = 2 * src->g - (src+2)->g;
		dest->b = 2 * src->b - (src+2)->b;
		dest->a = 2 * src->a - (src+2)->a;

		// left -1
		(dest+1)->r = 2 * src->r - (src + 1)->r;
		(dest+1)->g = 2 * src->g - (src + 1)->g;
		(dest+1)->b = 2 * src->b - (src + 1)->b;
		(dest+1)->a = 2 * src->a - (src + 1)->a;
		
		
		// right +1
		(dest+w+(BORDERSIZE/2))->r = 2 * (src+w-1)->r - (src+w-2)->r;
		(dest+w+(BORDERSIZE/2))->g = 2 * (src+w-1)->g - (src+w-2)->g;
		(dest+w+(BORDERSIZE/2))->b = 2 * (src+w-1)->b - (src+w-2)->b;
		(dest+w+(BORDERSIZE/2))->a = 2 * (src+w-1)->a - (src+w-2)->a;

		// right +2
		(dest+w+BORDERSIZE-1)->r = 2 * (src+w-1)->r - (src+w-3)->r;
		(dest+w+BORDERSIZE-1)->g = 2 * (src+w-1)->g - (src+w-3)->g;
		(dest+w+BORDERSIZE-1)->b = 2 * (src+w-1)->b - (src+w-3)->b;
		(dest+w+BORDERSIZE-1)->a = 2 * (src+w-1)->a - (src+w-3)->a;
		src += w + BORDERSIZE;
		dest += w + BORDERSIZE;
	}

	// new width is +BORDERSIZE
	w += BORDERSIZE;

	src = dest = temp;
	src += (w * 2);
	// copy/reflect pixels around top and bottom
	for (x=0; x<w; x++)
	{
		// top line - 2
		dest->r = 2 * src->r - (src+(w*2))->r;
		dest->g = 2 * src->g - (src+(w*2))->g;
		dest->b = 2 * src->b - (src+(w*2))->b;
		dest->a = 2 * src->a - (src+(w*2))->a;

		// top line - 1 
		(dest+w)->r = 2 * src->r - (src+w)->r;
		(dest+w)->g = 2 * src->g - (src+w)->g;
		(dest+w)->b = 2 * src->b - (src+w)->b;
		(dest+w)->a = 2 * src->a - (src+w)->a;

		// bottom line + 1
		(dest+(w*(h+BORDERSIZE-2)))->r = 2 * (src+(w*(h-1)))->r;
		(dest+(w*(h+BORDERSIZE-2)))->g = 2 * (src+(w*(h-1)))->g;
		(dest+(w*(h+BORDERSIZE-2)))->b = 2 * (src+(w*(h-1)))->b;
		(dest+(w*(h+BORDERSIZE-2)))->a = 2 * (src+(w*(h-1)))->a;

		// bottom line + 2
		(dest+(w*(h+BORDERSIZE-1)))->r = 2 * (src+(w*(h-2)))->r;
		(dest+(w*(h+BORDERSIZE-1)))->g = 2 * (src+(w*(h-2)))->g;
		(dest+(w*(h+BORDERSIZE-1)))->b = 2 * (src+(w*(h-2)))->b;
		(dest+(w*(h+BORDERSIZE-1)))->a = 2 * (src+(w*(h-2)))->a;

		dest++;
		src++;
	}

	// whew, return the bitmap
	return temp;
}

//=====================================================================
// Function:	Reduce
// Description:	Create a half sized image from a gaussian image
// Parameters:	src = pointer to INTRGBA data
//				w = width of image
//				h = height of image
// Returns:		pointer to new reduced INTRGBA data
//=====================================================================
INTRGBA *Reduce(INTRGBA *src, int w, int h)
{
	INTRGBA *dest;
	INTRGBA *newSrc;
	INTRGBA *temp;
	int x,y;

	newSrc = ConvolveImage(src, &GaussianBlurKernel, w, h);
	if (newSrc == NULL)
	{
		return NULL;
	}

	dest = new INTRGBA[(w/2) * (h/2)];
	if (dest == NULL)
	{
		MessageBox(NULL, _T("Unable to create temporary bitmap for reduction.\n"),
			_T("Memory Error!"), MB_OK);
		return NULL;
	}

	temp = dest;
	src = newSrc;
	for (y=0; y<h/2; y++)
	{
		for (x=0; x<w/2; x++)
		{
			*dest++ = *newSrc++;
			newSrc += 1;
		}
		newSrc += w;
	}

	delete src;
	return temp;
}

//=====================================================================
// Function:	Expand
// Description:	Create a double sized image and run gaussian kernal on
//				it
// Parameters:	src = pointer to INTRGBA data
//				w = width of image
//				h = height of image
// Returns:		pointer to newly expande INTRGBA data
//=====================================================================
INTRGBA *Expand(INTRGBA *src, int w, int h)
{
	INTRGBA *dest1;
	INTRGBA *dest2;
	INTRGBA *newDest;
	INTRGBA *temp;
	int x,y;

	dest1 = new INTRGBA[(w*2) * (h*2)];
	if (dest1 == NULL)
	{
		MessageBox(NULL, _T("Unable to create temporary bitmap for expansion.\n"),
			_T("Memory Error!"), MB_OK);
		return NULL;
	}

	temp = dest2 = dest1;
	dest2 += (w * 2);

	for (y=0; y<h; y++)
	{
		for (x=0; x<w; x++)
		{
			*dest1++ = *src;
			*dest2++ = *src;
			*dest1++ = *src;
			*dest2++ = *src++;
		}
		dest1 += (w * 2);
		dest2 += (w * 2);
	}

	newDest = ConvolveImage(temp, &GaussianBlurKernel, w*2, h*2);
	if (newDest == NULL)
	{
		return NULL;
	}

	delete temp;
	return newDest;
}

//=====================================================================
// Function:	Subtract
// Description:	Subtracts two images from one another
// Parameters:	src1 = first pointer to INTRGBA data
//				src2 = second pointer to INTRGBA data
//				w = width of image
//				h = height of image
// Returns:		pointer to newly subtracted INTRGBA data
//=====================================================================
INTRGBA *Subtract(INTRGBA *src1, INTRGBA *src2, int w, int h)
{
	INTRGBA *newDest;
	INTRGBA *temp;
	int x,y;
	int r, g, b, a;

	newDest = new INTRGBA[w * h];
	if (newDest == NULL)
	{
		MessageBox(NULL, _T("Unable to create temporary bitmap for subtraction.\n"),
			_T("Memory Error!"), MB_OK);
		return NULL;
	}

	temp = newDest;
	for (y=0; y<h; y++)
	{
		for (x=0; x<w; x++)
		{
			r = src1->r - src2->r;
			g = src1->g - src2->g;
			b = src1->b - src2->b;
			a = src1->a - src2->a;
#if 0
			// clamp?
			r = min(r, 65535);
			r = max(r, 0);
			g = min(g, 65535);
			g = max(g, 0);
			b = min(b, 65535);
			b = max(b, 0);
#endif
			newDest->r = r;
			newDest->g = g;
			newDest->b = b;
			newDest->a = a;

			newDest++;
			src1++;
			src2++;
		}
	}

	return temp;
}


//=====================================================================
// Function:	ScaleIntensity
// Description:	Scale intensity of image by number
// Parameters:	src1 = first pointer to INTRGBA data
//				w = width of image
//				h = height of image
// Returns:		.
//=====================================================================
void ScaleIntensity(INTRGBA *src1, int w, int h, int value)
{
	int x, y;
	for (y=0; y<h; y++)
	{
		for (x=0; x<w; x++)
		{
			src1->r = src1->r * value;
			src1->g = src1->g * value;
			src1->b = src1->b * value;
			src1->a = src1->a * value;
			src1++;
		}
	}
}

//=====================================================================
// Function:	Add
// Description:	Adds two images to each other
// Parameters:	src1 = first pointer to INTRGBA data
//				src2 = second pointer to INTRGBA data
//				w = width of image
//				h = height of image
// Returns:		pointer to newly added INTRGBA data
//=====================================================================
INTRGBA *Add(INTRGBA *src1, INTRGBA *src2, int w, int h)
{
	INTRGBA *newDest;
	INTRGBA *temp;
	int x,y;
	int r, g, b, a;

	newDest = new INTRGBA[w * h];
	if (newDest == NULL)
	{
		MessageBox(NULL, _T("Unable to create temporary bitmap for addition.\n"),
			_T("Memory Error!"), MB_OK);
		return NULL;
	}

	temp = newDest;
	for (y=0; y<h; y++)
	{
		for (x=0; x<w; x++)
		{
			r = src1->r + src2->r;
			g = src1->g + src2->g;
			b = src1->b + src2->b;
			a = src1->a + src2->a;
#if 0
			// clamp?
			r = min(r, 65535);
			r = max(r, 0);
			g = min(g, 65535);
			g = max(g, 0);
			b = min(b, 65535);
			b = max(b, 0);
#endif
			newDest->r = r;
			newDest->g = g;
			newDest->b = b;
			newDest->a = a;

			newDest++;
			src1++;
			src2++;
		}
	}

	return temp;
}

//=====================================================================
// Function:	ConvolveImage
// Description:	Use kernel to convolve image
// Parameters:	Bitmap = 64 bit TRUE color bitmap pointer
//				kernel = 5x5 integer matrix of weights
//				w = width of image
//				h = height of image
// Returns:		pointer to new image or NULL
//=====================================================================
INTRGBA *ConvolveImage(INTRGBA *bitmap, int (*kernel)[5][5], int w, int h)
{
	int i, j;
	int x, y;
	INTRGBA *src;
	INTRGBA *dest;
	INTRGBA *tempReturn;
	INTRGBA sum;
	int divisor;
	INTRGBA *tempBitmap;

	tempBitmap = CreateBorderImage(bitmap, w, h);

	if (tempBitmap == NULL)
	{
		MessageBox(NULL, _T("Unable to create temporary bitmap for filtering.\n"),
			_T("Memory Error!"), MB_OK);
		return NULL;
	}

	src = tempBitmap;
	dest = new INTRGBA[w * h];
	if (dest == NULL)
	{
		MessageBox(NULL, _T("Unable to create temporary bitmap for filtering.\n"),
			_T("Memory Error!"), MB_OK);
		return NULL;
	}

	tempReturn = dest;
	for (y=0; y<h; y++)
	{
		for (x=0; x<w; x++)
		{
			sum.r = 0;
			sum.g = 0;
			sum.b = 0;
			sum.a = 0;
			divisor = 0;

			INTRGBA *srcFilter;
			srcFilter = src;
			for (i=0; i<5; i++)
			{
				for (j=0; j<5; j++)
				{
					divisor += (*kernel)[i][j];
					sum.r += srcFilter->r * (*kernel)[i][j];
					sum.g += srcFilter->g * (*kernel)[i][j];
					sum.b += srcFilter->b * (*kernel)[i][j];
					sum.a += srcFilter->a * (*kernel)[i][j];
					srcFilter++;
				}
				srcFilter += w + BORDERSIZE - 5;
			}

			sum.r /= divisor;
			sum.g /= divisor;
			sum.b /= divisor;
			sum.a /= divisor;

			// clamp to 16 bits
			sum.r = min(sum.r, 65535);
			sum.r = max(sum.r, 0);
			sum.g = min(sum.g, 65535);
			sum.g = max(sum.g, 0);
			sum.b = min(sum.b, 65535);
			sum.b = max(sum.b, 0);
			sum.a = min(sum.a, 65535);
			sum.a = max(sum.a, 0);

			dest->r = sum.r;
			dest->g = sum.g;
			dest->b = sum.b;
			dest->a = sum.a;
			dest++;

			src++;
		}
		src += BORDERSIZE;
	}

	delete tempBitmap;
	return tempReturn;
}

typedef struct WHSIZE
{
	int	w;
	int h;
} WHSIZE;

//=====================================================================
// Function:	SplineImages
// Description:	Build Laplacian pyramids for splining images together
//				then splines the images together
// Parameters:	bitmap1 = first 64 bit TRUE color bitmap pointer
//				bitmap2 = second 64 bit TRUE color bitmap pointer
//				alpha contains image ID's
// Returns:		TRUE = success
//=====================================================================
BOOL SplineImages(Bitmap *bitmap1, Bitmap *bitmap2)
{
	INTRGBA *reduced1, *reduced2;
	INTRGBA *expanded;
	INTRGBA *Laplacian[2][8];
	INTRGBA *Gaussian[2][8];
	INTRGBA *temp;
	INTRGBA *srcTemp;
	BMM_Color_64 *line;
	BMM_Color_64 *tempLine;
	WHSIZE wh[8];
	int	pyramidLevels;

	int j, k;
	int x, y;
	int w, h;
	int tw, th;

	w = bitmap1->Width();
	h = bitmap1->Height();

	if ((w < 1) || (h < 1))
	{
		MessageBox(NULL, _T("Invalid size specified for bitmap.\n"),
			_T("Dimension Error!"), MB_OK);
		return NULL;
	}

	line = new BMM_Color_64[w];
	if (line == NULL)
	{
		MessageBox(NULL, _T("Unable to create temporary bitmap line for Gaussian filter.\n"),
			_T("Memory Error!"), MB_OK);
		return NULL;
	}

	// first allocate temporary storage
	Gaussian[0][0] = new INTRGBA[w * h];
	if (Gaussian[0][0] == NULL)
	{
		delete line;
		MessageBox(NULL, _T("Unable to create temporary bitmap for Gaussian filter.\n"),
			_T("Memory Error!"), MB_OK);
		return NULL;
	}

	temp = Gaussian[0][0];
	for (y=0; y<h; y++)
	{
		bitmap1->GetPixels(0, y, w, line);
		tempLine = line;
		for (x=0; x<w; x++)
		{
			// convert to 32 bit integers
			temp->r = (int)tempLine->r;
			temp->g = (int)tempLine->g;
			temp->b = (int)tempLine->b;
			temp->a = 0; //(int)tempLine->a;
			temp++;
			tempLine++;
		}
	}

	Gaussian[1][0] = new INTRGBA[w*h];
	if (Gaussian[1][0] == NULL)
	{
		delete line;
		delete Gaussian[0][0];

		MessageBox(NULL, _T("Unable to create temporary bitmap for temporary buffer.\n"),
			_T("Memory Error!"), MB_OK);
		return NULL;
	}

	temp = Gaussian[1][0];
	for (y=0; y<h; y++)
	{
		bitmap2->GetPixels(0, y, w, line);
		tempLine = line;
		for (x=0; x<w; x++)
		{
			// convert to 32 bit integers
			temp->r = (int)tempLine->r;
			temp->g = (int)tempLine->g;
			temp->b = (int)tempLine->b;
			temp->a = (int)tempLine->a;
			temp++;
			tempLine++;
		}
	}

	reduced1 = Gaussian[0][0];
	reduced2 = Gaussian[1][0];
	tw = w;
	th = h;
	pyramidLevels = 0;
	for (j=0; j< 7; j++)
	{
		// next reduction will cause pixel loss, so stop here
		if ((tw & 1) || (th & 1))
			break;

		wh[j].w = tw;
		wh[j].h = th;
		Gaussian[0][j] = reduced1;
		reduced1 = Reduce(Gaussian[0][j], tw, th);
		if (reduced1 == NULL)
		{
			delete Gaussian[0][j];
			delete line;
			for (k=0; k<j; k++)
			{
				delete Gaussian[0][k];
				delete Gaussian[1][k];
				delete Laplacian[0][k];
				delete Laplacian[1][k];
			}
			return FALSE;
		}

		expanded = Expand(reduced1, tw/2, th/2);
		if (expanded == NULL)
		{
			delete reduced1;
			delete line;
			delete Gaussian[0][j];
			for (k=0; k<j; k++)
			{
				delete Gaussian[0][k];
				delete Gaussian[1][k];
				delete Laplacian[0][k];
				delete Laplacian[1][k];
			}
			return FALSE;
		}

		Laplacian[0][j] = Subtract(Gaussian[0][j], expanded, tw, th);
		delete expanded;
		if (Laplacian[0][j] == NULL)
		{
			delete reduced1;
			delete Gaussian[0][j];
			delete line;
			for (k=0; k<j; k++)
			{
				delete Gaussian[0][k];
				delete Gaussian[1][k];
				delete Laplacian[0][k];
				delete Laplacian[1][k];
			}
			return FALSE;
		}

		Gaussian[1][j] = reduced2;
		reduced2 = Reduce(Gaussian[1][j], tw, th);
		if (reduced2 == NULL)
		{
			delete reduced1;
			delete line;
			delete Gaussian[0][j];
			delete Gaussian[1][j];
			for (k=0; k<j; k++)
			{
				delete Gaussian[0][k];
				delete Gaussian[1][k];
				delete Laplacian[0][k];
				delete Laplacian[1][k];
			}
			return FALSE;
		}

		expanded = Expand(reduced2, tw/2, th/2);
		if (expanded == NULL)
		{
			delete reduced1;
			delete reduced2;
			delete Gaussian[0][j];
			delete Gaussian[1][j];
			delete line;
			for (k=0; k<j; k++)
			{
				delete Gaussian[0][k];
				delete Gaussian[1][k];
				delete Laplacian[0][k];
				delete Laplacian[1][k];
			}
			return FALSE;
		}

		Laplacian[1][j] = Subtract(Gaussian[1][j], expanded, tw, th);
		delete expanded;
		if (Laplacian[1][j] == NULL)
		{
			delete reduced1;
			delete reduced2;
			delete Gaussian[0][j];
			delete Gaussian[1][j];
			delete line;
			for (k=0; k<j; k++)
			{
				delete Gaussian[0][k];
				delete Gaussian[1][k];
				delete Laplacian[0][k];
				delete Laplacian[1][k];
			}
			return FALSE;
		}

		pyramidLevels++;

		tw /= 2;
		th /= 2;
		if ((tw <= 1) || (th <= 1))
			break;

	}

	// last laplacian is based on last gaussian (reduced image)
	wh[pyramidLevels].w = tw;
	wh[pyramidLevels].h = th;
	Laplacian[0][pyramidLevels] = Gaussian[0][pyramidLevels] = reduced1;
	Laplacian[1][pyramidLevels] = Gaussian[1][pyramidLevels] = reduced2;
	pyramidLevels++;

	// we now have all the pyramids built, form a combined one based on alpha mask
	for (j=0; j<pyramidLevels; j++)
	{
		temp = Gaussian[0][j];	// destination
		srcTemp = Gaussian[1][j];	// source alpha
		reduced1 = Laplacian[0][j];
		reduced2 = Laplacian[1][j];
		// combine pyramid images using alpha
		for (y=0; y<wh[j].h; y++)
		{
			for (x=0; x<wh[j].w; x++)
			{
				int alpha;
				alpha = srcTemp->a;
				temp->r = (((__int64)reduced2->r * alpha) + ((__int64)reduced1->r * (65535-alpha))) 
							/ 65535;
				temp->g = (((__int64)reduced2->g * alpha) + ((__int64)reduced1->g * (65535-alpha))) 
							/ 65535;
				temp->b = (((__int64)reduced2->b * alpha) + ((__int64)reduced1->b * (65535-alpha))) 
							/ 65535;
				temp->a = 65535;

				srcTemp++;
				temp++;
				reduced1++;
				reduced2++;
			}
		}
		// delete unneed Laplacian/Gaussian pyramids, since these were combined into 
		// Gaussian slots[0][j]
		delete Gaussian[1][j];
		// since last Laplacian == last Gaussian, don't delete it again
		if (j != pyramidLevels-1)
		{
			delete Laplacian[0][j];
			delete Laplacian[1][j];
		}
	}

	// expand and sum the pyramids
	for (j=pyramidLevels-1; j>=1; j--)
	{
		expanded = Expand(Gaussian[0][j], wh[j].w, wh[j].h);
		if (expanded == NULL)
		{
			delete line;
			for (k=0; k<=j; k++)
			{
				delete Gaussian[0][k];
				return FALSE;
			}
		}
		// increase level
		ScaleIntensity(expanded, wh[j-1].w, wh[j-1].h, 1);
		// delete unused gaussian entry
		delete  Gaussian[0][j];
		reduced1 = Add(Gaussian[0][j-1], expanded, wh[j-1].w, wh[j-1].h);
		if (reduced1 == NULL)
		{
			delete line;
			for (k=0; k<j; k++)
			{
				delete Gaussian[0][k];
				return FALSE;
			}
		}
		delete expanded;
		delete Gaussian[0][j-1];
		Gaussian[0][j-1] = reduced1;

	}

#if 1
	temp = Gaussian[0][0];
	for (y=0; y<h; y++)
	{
		tempLine = line;
		for (x=0; x<w; x++)
		{
			INTRGBA clampRGBA;
			clampRGBA.r = min(temp->r, 65535);
			clampRGBA.r = max(temp->r, 0);
			clampRGBA.g = min(temp->g, 65535);
			clampRGBA.g = max(temp->g, 0);
			clampRGBA.b = min(temp->b, 65535);
			clampRGBA.b = max(temp->b, 0);
			clampRGBA.a = min(temp->a, 65535);
			clampRGBA.a = max(temp->a, 0);
			tempLine->r = clampRGBA.r;
			tempLine->g = clampRGBA.g;
			tempLine->b = clampRGBA.b;
			tempLine->a = clampRGBA.a;
			temp++;
			tempLine++;
		}

		bitmap1->PutPixels(0, y, w, line);
	}
#endif

	delete line;
	delete Gaussian[0][0];
	return TRUE;
}

